import logging


def create_console_handler():
    console_handler = logging.StreamHandler()
    console_formatter = logging.Formatter(
        "%(asctime)s %(name)s %(levelname)s: %(message)s"
    )
    console_handler.setFormatter(console_formatter)
    return console_handler


def get_logger_name(clazz):
    return f"{clazz.__module__}.{clazz.__name__}"



import adafruit_sgp30
import busio

from klimalogger import Config, DataBuilder
from klimalogger.measurement import Measurements

from . import BaseSensor


class SGP30Sensor(BaseSensor):
    name = "SGP30"
    priority = 3

    def __init__(self, i2c_bus: busio.I2C, config: Config):
        self.baseline_eCO2 = config.baselines.get("eCO2")
        self.baseline_TVOC = config.baselines.get("TVOC")

        self.driver = adafruit_sgp30.Adafruit_SGP30(i2c_bus)

    def measure(self, data_builder: DataBuilder, _: Measurements) -> None:
        # self.sensor.set_iaq_baseline(self.baseline_eCO2, self.baseline_TVOC)
        carbon_dioxide_equivalent, total_volatile_organic_compounds = (
            self.driver.iaq_measure()
        )

        data_builder.add(self.name, "eCO2", "ppm", float(carbon_dioxide_equivalent))
        data_builder.add(
            self.name, "TVOC", "ppb", float(total_volatile_organic_compounds)
        )



import json
import logging
import secrets

from paho.mqtt import client as mqtt_client
from paho.mqtt.enums import CallbackAPIVersion

# StoreClient removed; using concrete class
from klimalogger import config

log = logging.getLogger(__name__)


class QueueTransport:
    def __init__(self, configuration: config.Config):
        self.qos = configuration.mqtt_qos
        self.mqtt_prefix = configuration.mqtt_prefix
        self.client: mqtt_client.Client | None = None

        client_id = (
            f"klimalogger-mqtt-{configuration.host_name}-{secrets.randbelow(1000)}"
        )

        def on_connect(client, userdata, flags, reason_code, properties):
            if reason_code == 0:
                log.debug("Connected to MQTT Broker!")
            else:
                log.error("Failed to connect, return code %d", reason_code)

        def on_disconnect(client, userdata, flags, reason_code, properties):
            log.warning(f"Disconnected from MQTT Broker: {reason_code}")

        try:
            self.client = mqtt_client.Client(
                client_id=client_id,
                clean_session=False,
                callback_api_version=CallbackAPIVersion.VERSION2,
            )
            # client.username_pw_set(username, password)
            self.client.on_connect = on_connect
            self.client.on_disconnect = on_disconnect
            log.info(
                "connect to host %s, port %d",
                configuration.mqtt_host,
                configuration.mqtt_port,
            )
            self.client.connect(configuration.mqtt_host, configuration.mqtt_port)
            self.client.loop_start()
        except Exception:
            log.exception("could not create client")
            self.client = None

    def __del__(self):
        if self.client:
            self.client.disconnect()

    def store(self, data: list[dict]):
        if self.client:

            if not self.client.is_connected():
                log.warning("client not connected, try to reconnect")
                try:
                    self.client.reconnect()
                except Exception:
                    log.warning("reconnect failed")

            for entry in data:
                topic, json_message = self.map_entry(entry)
                message = json.dumps(json_message)
                log.info("write data (%d bytes) to topic %s", len(message), topic)
                result = self.client.publish(
                    topic,
                    payload=message,
                    qos=self.qos,
                )
                status = result[0]
                if status != 0:
                    log.warning(
                        "Failed to send message to topic %s with status %s",
                        topic,
                        status,
                    )
        else:
            log.warning("client not available")
            raise RuntimeError("bla")

    def map_entry(self, entry: dict):
        timestamp = entry["time"]
        value = entry["fields"]["value"]
        tags = entry["tags"]
        measurement_type = tags["type"]
        unit = tags["unit"]
        sensor = tags["sensor"]
        topic = f"{self.mqtt_prefix}/{measurement_type}"
        return (
            topic,
            {
                "time": int(timestamp),
                "value": value,
                "unit": unit,
                "sensor": sensor,
                "calculated": tags["calculated"],
            },
        )



import adafruit_bh1750
import busio

from ..data_builder import DataBuilder
from ..measurement import Measurements


class BH1750Sensor:
    name = "BH1750"
    priority = 1

    def __init__(self, i2c_bus: busio.I2C, address: int):
        self.driver = adafruit_bh1750.BH1750(i2c_bus, address)

    def __del__(self):
        self.driver.stop_periodic_measurement()

    def measure(self, data_builder: DataBuilder, measurements: Measurements) -> None:
        light = self.driver.lux

        data_builder.add(self.name, "light", "Lux", float(light))



from __future__ import annotations

import logging
from dataclasses import dataclass

log = logging.getLogger(__name__)


@dataclass
class Measurements:
    temperature: float | None = None
    relative_humidity: float | None = None



import adafruit_veml7700
import busio

from ..data_builder import DataBuilder
from ..measurement import Measurements


class VEML7700Sensor:
    name = "VEML7700"
    priority = 1

    def __init__(self, i2c_bus: busio.I2C, address: int):
        self.driver = adafruit_veml7700.VEML7700(i2c_bus, address)
        self.driver.data_rate = 10  # in Hz, from 1-255 or 1000
        self.driver.continuous_mode = True

    def __del__(self):
        self.driver.stop_periodic_measurement()

    def measure(self, data_builder: DataBuilder, measurements: Measurements) -> None:
        light = self.driver.light

        data_builder.add(self.name, "light", "Lux", float(light))



import adafruit_sgp40
import busio

from klimalogger import DataBuilder
from klimalogger.measurement import Measurements

from . import BaseSensor


class SGP40Sensor(BaseSensor):
    name = "SGP40"
    priority = 3

    def __init__(self, i2c_bus: busio.I2C, address: int):
        self.driver = adafruit_sgp40.SGP40(i2c_bus, address)

    def measure(self, data_builder: DataBuilder, measurements: Measurements) -> None:
        if (
            measurements.temperature is not None
            and measurements.relative_humidity is not None
        ):
            data_builder.add(
                self.name,
                "VOC index",
                "",
                float(
                    self.driver.measure_index(
                        temperature=measurements.temperature,
                        relative_humidity=measurements.relative_humidity,
                    )
                ),
            )
        else:
            value = self.driver.raw
            data_builder.add(self.name, "raw gas", "", float(value))



import configparser
import logging
import os
import socket
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class Config:
    mqtt_host: str
    mqtt_port: int
    mqtt_prefix: str
    mqtt_qos: int = 1
    mqtt_username: str | None = None
    mqtt_password: str | None = None
    host_name: str = socket.gethostname()
    sensors: list[int] | None = None
    elevation: int | None = None
    baselines: dict[str, float] = field(default_factory=dict)
    device_map: dict[int, str] = field(default_factory=dict)


def is_circuitpython():
    return sys.platform == "circuitpython"


def ensure_not_none(value: Any | None, name: str = ""):
    if value is None:
        raise ValueError(f"{name} is required")
    return value


def build_config() -> Config:
    if is_circuitpython():
        return build_env_based_config()
    else:
        cfg = build_file_based_config()
        # Ensure we always return a valid Config instance
        if not isinstance(cfg, Config):
            raise RuntimeError("Invalid configuration")
        return cfg


def device_map(value: str) -> dict[int, str]:
    return {
        int(elements[0]): elements[1]
        for entry in value.split(",")
        if len(elements := entry.split("=")) > 1
    }


def sensors(value: str | None) -> list[int] | None:
    return (
        [int(address, 16) for address in value.split(",")]
        if value is not None
        else None
    )


def build_env_based_config():
    return Config(
        mqtt_host=ensure_not_none(os.getenv("MQTT_HOST")),
        mqtt_port=int(os.getenv("MQTT_PORT")),
        mqtt_prefix=os.getenv("MQTT_PREFIX"),
        mqtt_username=os.getenv("MQTT_USERNAME", None),
        mqtt_password=os.getenv("MQTT_PASSWORD", None),
        elevation=int(os.getenv("ELEVATION", "0")),
        device_map=device_map(os.getenv("DEVICE_MAP", "")),
    )


def build_file_based_config():
    config_parser = load_config_parser()

    return Config(
        mqtt_host=ensure_not_none(config_parser.get("queue", "host"), "queue host"),
        mqtt_port=int(config_parser.get("queue", "port")),
        mqtt_prefix=config_parser.get("queue", "queue_prefix", fallback="sensors"),
        mqtt_qos=int(config_parser.get("queue", "queue_qos", fallback="1")),
        mqtt_username=config_parser.get("queue", "username", fallback=None),
        mqtt_password=config_parser.get("queue", "password", fallback=None),
        elevation=int(config_parser.get("client", "elevation", fallback="0")),
        sensors=sensors(config_parser.get("client", "sensors", fallback=None)),
        device_map=device_map(config_parser.get("client", "device_map", fallback="")),
    )


def load_config_parser() -> configparser.ConfigParser:
    """Load the configuration from standard locations, replacing the Injector-based provider."""
    etc = Path("/etc")
    config_filename = "klimalogger.conf"

    config_file_locations = [
        Path(config_filename),
        etc / "klimalogger" / config_filename,
        etc / config_filename,
    ]

    for config_file_location in config_file_locations:
        if config_file_location.exists():
            log.info("reading config file location %s", config_file_location)
            config_parser = configparser.ConfigParser()
            config_parser.read(config_file_location)
            return config_parser

    raise OSError("config file not found")



import math

import adafruit_mmc56x3
import busio

from ..data_builder import DataBuilder
from ..measurement import Measurements


class MMC56x3Sensor:
    name = "MMC56x3"
    priority = 1

    def __init__(self, i2c_bus: busio.I2C, address: int):
        self.driver = adafruit_mmc56x3.MMC5603(i2c_bus, address)
        # self.driver.data_rate = 10  # in Hz, from 1-255 or 1000
        # self.driver.continuous_mode = True

    def measure(self, data_builder: DataBuilder, measurements: Measurements) -> None:
        mag_x, mag_y, mag_z = self.driver.magnetic

        data_builder.add(self.name, "magX", "uT", float(mag_x))
        data_builder.add(self.name, "magY", "uT", float(mag_y))
        data_builder.add(self.name, "magZ", "uT", float(mag_z))

        mag = math.sqrt(mag_x**2 + mag_y**2 + mag_z**2)
        data_builder.add(self.name, "mag", "uT", float(mag), is_calculated=True)



import logging
import time

from . import logger
from .config import Config, build_config
from .data_builder import DataBuilder
from .sensors import Sensors
from .transport import QueueTransport

root_logger = logging.getLogger(__name__)
root_logger.setLevel(logging.WARN)

log = root_logger


def set_log_level(log_level):
    root_logger.setLevel(log_level)


def add_log_handler(log_handler):
    root_logger.addHandler(log_handler)


def set_parent_logger(logger):
    logger.parent = root_logger


class Client:
    def __init__(self, sensors: Sensors, transport: QueueTransport):
        self.sensors = sensors
        self.transport = transport

    def measure_and_store_periodically(self, period=15):
        log.info("measure_and_store_periodically(%d)", period)
        last_measurement = time.time() - period
        skip_initial = 2
        while True:
            timestamp = time.time()
            if last_measurement < timestamp - period:
                if skip_initial == 0:
                    self.measure_and_store()
                else:
                    log.info("skipping initial measurement(%d)", skip_initial)
                    self.measure()
                    skip_initial -= 1
                last_measurement += period
            else:
                self.measure()
            log.info("time taken: %d", time.time() - last_measurement)
            time.sleep(1)

    def measure_and_store(self):
        log.info(
            "measure_and_store()",
        )
        data_builder = self.measure()

        self.store_data(data_builder.data)

    def measure(self):
        return self.sensors.measure()

    def store_data(self, data):
        try:
            self.transport.store(data)
            log.info("stored data")
        except Exception as e:
            log.error(
                "error during data transmission: create local log entry", exc_info=e
            )


def client():
    import board

    # Manual wiring instead of Injector
    config = build_config()
    sensors = Sensors(config, board.I2C())
    store = QueueTransport(config)
    return Client(sensors, store)


__all__ = ["client", "Config", "DataBuilder", "logger"]



import logging

import adafruit_bme680
import busio

from .. import DataBuilder
from ..calc import PressureCalc, TemperatureCalc
from ..config import Config
from ..measurement import Measurements
from . import BaseSensor

log = logging.getLogger(__name__)


class BME680Sensor(BaseSensor):
    name = "BME680"
    priority = 1

    def __init__(
        self,
        i2c_bus: busio.I2C,
        address: int,
        config: Config,
        temperature_calc: TemperatureCalc,
        pressure_calc: PressureCalc,
    ):
        log.info("init()")
        self.temperature_calc = temperature_calc
        self.pressure_calc = pressure_calc
        self.elevation = int(config.elevation or 0)
        self.driver = adafruit_bme680.Adafruit_BME680_I2C(i2c_bus, address)
        self.driver.set_gas_heater(None, None)

    def measure(self, data_builder: DataBuilder, measurements: Measurements) -> None:
        temperature = self.driver.temperature
        relative_humidity = self.driver.relative_humidity
        dew_point = self.temperature_calc.dew_point(temperature, relative_humidity)
        pressure = self.driver.pressure
        sea_level_pressure = self.pressure_calc.sea_level_pressure(
            pressure, temperature, self.elevation
        )
        voc_gas = self.driver.gas

        data_builder.add(self.name, "temperature", "°C", round(temperature, 2))
        data_builder.add(
            self.name, "relative humidity", "%", round(relative_humidity, 2)
        )
        data_builder.add(
            self.name, "dew point", "°C", round(dew_point, 2), is_calculated=True
        )
        data_builder.add(self.name, "pressure", "hPa", round(pressure, 2))
        data_builder.add(
            self.name, "sea level pressure", "hPa", round(sea_level_pressure, 2)
        )
        data_builder.add(self.name, "voc gas", "Ohm", float(voc_gas))

        measurements.temperature = temperature
        measurements.relative_humidity = relative_humidity



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import busio

if TYPE_CHECKING:
    # Imported only for type checking to avoid circular imports at runtime
    from ..data_builder import DataBuilder
    from ..measurement import Measurements


class BaseSensor(ABC):
    """Common base class for all sensors.

    Provides a shared interface and common attributes that all sensors expose.
    Subclasses must implement the `measure` method and should define `name`
    and `priority` class attributes.
    """

    # Human-readable sensor name; subclasses should override
    name: str = "Sensor"
    # Lower numbers run earlier; subclasses should override
    priority: int = 100

    @abstractmethod
    def measure(
        self, data_builder: "DataBuilder", measurements: "Measurements"
    ) -> None:
        """Perform a measurement and add values via the DataBuilder.

        Implementations may also read/update `measurements` to share context
        (e.g., temperature/humidity) across sensors.
        """
        raise NotImplementedError


def create_i2c_bus() -> busio.I2C:
    import board

    return busio.I2C(board.SCL, board.SDA, frequency=100000)



import time


class DataBuilder:
    def __init__(self):
        self.timestamp = round(time.time(), 1)
        self.data = []

    def __repr__(self):
        return f"@{self.timestamp}: {self.data}"

    def add(
        self,
        sensor: str,
        measurement_type: str,
        measurement_unit: str,
        measurement_value: float,
        is_calculated: bool = False,
    ):
        if measurement_value is not None:
            self.data += [
                self.create(
                    sensor,
                    measurement_type,
                    measurement_unit,
                    measurement_value,
                    is_calculated,
                )
            ]

    def create(
        self,
        sensor: str,
        measurement_type: str,
        measurement_unit: str,
        measurement_value: float,
        is_calculated: bool = False,
    ) -> dict:
        return {
            "measurement": "data",
            "tags": {
                "type": measurement_type,
                "unit": measurement_unit,
                "sensor": sensor,
                "calculated": is_calculated,
            },
            "time": self.timestamp,
            "fields": {"value": measurement_value},
        }



import adafruit_bmp3xx
import busio

from .. import DataBuilder
from ..calc import PressureCalc
from ..config import Config
from ..measurement import Measurements
from . import BaseSensor


class BMP3xxSensor(BaseSensor):
    name = "BMP3xx"
    priority = 2

    def __init__(
        self,
        i2c_bus: busio.I2C,
        address: int,
        config: Config,
        pressure_calc: PressureCalc,
    ):
        self.pressure_calc = pressure_calc
        self.elevation = int(config.elevation or 0)

        self.driver = adafruit_bmp3xx.BMP3XX_I2C(i2c_bus, address)

    def measure(self, data_builder: DataBuilder, _: Measurements) -> None:
        temperature = self.driver.temperature
        pressure = self.driver.pressure
        sea_level_pressure = self.pressure_calc.sea_level_pressure(
            pressure, temperature, self.elevation
        )

        data_builder.add(self.name, "pressure", "hPa", round(pressure, 2))
        data_builder.add(
            self.name,
            "sea level pressure",
            "hPa",
            round(sea_level_pressure, 2),
            is_calculated=True,
        )



import logging
import time
from collections.abc import Callable

import busio

from klimalogger import DataBuilder

from .calc import PressureCalc, TemperatureCalc
from .config import Config
from .measurement import Measurements
from .sensor import BaseSensor
from .sensor.bh1750 import BH1750Sensor
from .sensor.bme680 import BME680Sensor
from .sensor.bmp3xx import BMP3xxSensor
from .sensor.mmc56x3 import MMC56x3Sensor
from .sensor.scd4x import SCD4xSensor
from .sensor.sgp40 import SGP40Sensor
from .sensor.sht4x import SHT4xSensor
from .sensor.veml7700 import VEML7700Sensor

log = logging.getLogger(__name__)


def scan(i2c_bus: busio.I2C):
    lock_attempts = 0
    while not i2c_bus.try_lock():
        lock_attempts += 1

    if lock_attempts > 1:
        log.debug(f"scan() attempts: {lock_attempts}")

    devices = []
    iterations = 0
    for _ in range(20):
        devices = i2c_bus.scan()
        if devices:
            break
        time.sleep(0.2)
        iterations += 1
    log.info(f"Found {len(devices)} devices after {iterations} iterations.")

    i2c_bus.unlock()

    return devices


class Sensors:
    sensor_map: dict[str, Callable] = {
        SCD4xSensor.name: lambda i2c_bus, address, _: SCD4xSensor(i2c_bus, address),
        SGP40Sensor.name: lambda i2c_bus, address, _: SGP40Sensor(i2c_bus, address),
        SHT4xSensor.name: lambda i2c_bus, address, _: SHT4xSensor(
            i2c_bus, address, TemperatureCalc()
        ),
        BMP3xxSensor.name: lambda i2c_bus, address, config: BMP3xxSensor(
            i2c_bus, address, config, PressureCalc()
        ),
        BME680Sensor.name: lambda i2c_bus, address, config: BME680Sensor(
            i2c_bus, address, config, TemperatureCalc(), PressureCalc()
        ),
        MMC56x3Sensor.name: lambda i2c_bus, address, _: MMC56x3Sensor(i2c_bus, address),
        VEML7700Sensor.name: lambda i2c_bus, address, _: VEML7700Sensor(
            i2c_bus, address
        ),
        BH1750Sensor.name: lambda i2c_bus, address, _: BH1750Sensor(i2c_bus, address),
    }

    def __init__(self, config: Config, i2c_bus: busio.I2C):
        self.config = config
        self.i2c_bus = i2c_bus
        self.sensors: list[BaseSensor] = []

        self.device_map: dict[int, str] = {
            16: VEML7700Sensor.name,
            35: BH1750Sensor.name,
            48: MMC56x3Sensor.name,
            68: SHT4xSensor.name,
            89: SGP40Sensor.name,
            98: SCD4xSensor.name,
            119: BMP3xxSensor.name,
        }
        device_map = config.device_map
        if device_map:
            self.device_map.update(device_map)
        self.address_map: dict[str, int] = {
            sensor_name: address for address, sensor_name in self.device_map.items()
        }

        self.scan_devices()

    def scan_devices(self):
        sensors_in_use = {sensor.name for sensor in self.sensors}

        device_addresses = (
            scan(self.i2c_bus) if not self.config.sensors else self.config.sensors
        )

        sensors_found = {
            self.device_map[device_address]
            for device_address in device_addresses
            if device_address in self.device_map
        }
        unknown_sensors_found = {
            str(device_address)
            for device_address in device_addresses
            if device_address not in self.device_map
        }
        if unknown_sensors_found:
            log.warning(
                f"Could not find sensors for addresses: {', '.join(unknown_sensors_found)}"
            )

        if sensors_in_use != sensors_found:

            sensors = [
                sensor for sensor in self.sensors if sensor.name in sensors_found
            ]
            for sensor_name in sensors_found.difference(sensors_in_use):
                if sensor_name in self.sensor_map:
                    sensor = self.sensor_map[sensor_name]
                    address = self.address_map[sensor_name]
                    sensors.append(sensor(self.i2c_bus, address, self.config))

            sensors.sort(key=lambda sensor: sensor.priority)

            log.info("updated sensors:")
            for sensor in sensors:
                log.info(f"  {sensor.name} {sensor.priority}")

            self.sensors = sensors

    def measure(self, data_builder=None) -> DataBuilder:
        if data_builder is None:
            data_builder = DataBuilder()
        measurements = Measurements()

        for sensor in self.sensors:
            try:
                start_time = time.monotonic_ns()
                sensor.measure(data_builder, measurements)
                end_time = time.monotonic_ns()
                data_builder.add(
                    sensor.name, "time", "ms", (end_time - start_time) / 1e6
                )
            except OSError:
                pass

        return data_builder
